"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import pytest
import shephard
from shephard.interfaces import si_sites
from shephard.exceptions import ProteinException
from shephard.apis import uniprot  

test_data_dir = shephard.get_data('test_data')


def test_get_site_by_position():
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    pos = TS1.protein('O00470').get_sites_by_position(72)    
    assert len(pos[72]) == 1
    assert pos[72][0].site_type == 'Sumoylation'


    pos = TS1.protein('O00470').get_sites_by_position(72, wiggle=5)    
    assert len(pos[72]) == 1
    assert pos[72][0].site_type == 'Sumoylation'


    pos = TS1.protein('O00470').get_sites_by_position(72, wiggle=5, return_list=True)    
    assert len(pos) == 1
    assert pos[0].site_type == 'Sumoylation'



    pos = TS1.protein('O00470').get_sites_by_position(196)    
    assert len(pos[196]) == 1
    assert pos[196][0].site_type == 'Phosphoserine'


    pos = TS1.protein('O00470').get_sites_by_position(197, wiggle=5)    
    assert len(pos) == 2
    assert pos[196][0].site_type == 'Phosphoserine'


    pos = TS1.protein('O00470').get_sites_by_position(197, wiggle=5, return_list=True)    
    assert len(pos) == 2
    assert pos[0].site_type == 'Phosphoserine'
    assert pos[1].site_type == 'Phosphoserine'


    pos = TS1.protein('O00470').get_sites_by_position(100, wiggle=5, return_list=True)    
    assert len(pos) == 0



def test_get_site_by_range():
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    pos = TS1.protein('O00470').get_sites_by_range(55,79)    
    assert len(pos[72]) == 1
    assert pos[72][0].site_type == 'Sumoylation'


    pos = TS1.protein('O00470').get_sites_by_range(70,80, wiggle=5)    
    assert len(pos[72]) == 1
    assert pos[72][0].site_type == 'Sumoylation'


    pos = TS1.protein('O00470').get_sites_by_range(55,79, wiggle=5, return_list=True)    
    assert len(pos) == 1
    assert pos[0].site_type == 'Sumoylation'



    pos = TS1.protein('O00470').get_sites_by_range(150,250)    
    assert len(pos[196]) == 1
    assert pos[196][0].site_type == 'Phosphoserine'


    pos = TS1.protein('O00470').get_sites_by_range(150,250, wiggle=5)    
    assert len(pos) == 2
    assert pos[196][0].site_type == 'Phosphoserine'


    pos = TS1.protein('O00470').get_sites_by_range(150,250, wiggle=5, return_list=True)    
    assert len(pos) == 2
    assert pos[0].site_type == 'Phosphoserine'
    assert pos[1].site_type == 'Phosphoserine'

    pos = TS1.protein('O00470').get_sites_by_range(30,60, wiggle=5, return_list=True)    
    assert len(pos) == 0


def test_get_site_by_type():
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    
    pos = TS1.protein('O00470').get_sites_by_type('Phosphoserine')
    assert len(pos[196]) == 1
    assert pos[196][0].site_type == 'Phosphoserine'
    assert pos[198][0].site_type == 'Phosphoserine'

    pos = TS1.protein('O00470').get_sites_by_type('Phosphoserine', return_list=True)
    assert len(pos) == 2


    pos = TS1.protein('O00470').get_sites_by_type('PhosphoX')
    assert len(pos) == 0

    pos = TS1.protein('O00470').get_sites_by_type('PhosphoX', return_list=True)
    assert len(pos) == 0

    
    # add a site, check we can find it, remove it and check it's gone!
    TS1.protein('O00470').add_site(1, 'PhosphoX')
    
    pos = TS1.protein('O00470').get_sites_by_type('PhosphoX', return_list=True)
    assert len(pos) == 1

    TS1.protein('O00470').remove_site(pos[0])
    pos = TS1.protein('O00470').get_sites_by_type('PhosphoX', return_list=True)
    assert len(pos) == 0


def test_get_site_by_type_and_range():
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    
    pos = TS1.protein('O00470').get_sites_by_type_and_range('Phosphoserine', 190, 200)
    assert len(pos[196]) == 1
    assert pos[196][0].site_type == 'Phosphoserine'
    assert pos[198][0].site_type == 'Phosphoserine'

    pos = TS1.protein('O00470').get_sites_by_type_and_range('Phosphoserine', 190, 200, return_list=True)
    assert len(pos) == 2


    pos = TS1.protein('O00470').get_sites_by_type_and_range('PhosphoX', 1, 50)
    assert len(pos) == 0

    pos = TS1.protein('O00470').get_sites_by_type_and_range('PhosphoX', 1, 50, return_list=True)
    assert len(pos) == 0

    
    # add a site, check we can find it, remove it and check it's gone!
    TS1.protein('O00470').add_site(1, 'PhosphoX')
    
    pos = TS1.protein('O00470').get_sites_by_type_and_range('PhosphoX', 1, 50, return_list=True)
    assert len(pos) == 1

    TS1.protein('O00470').remove_site(pos[0])
    pos = TS1.protein('O00470').get_sites_by_type_and_range('PhosphoX', 1, 50, return_list=True)
    assert len(pos) == 0




    

    
    



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import pytest

names_list = ['O00401', 'O00470', 'O00472', 'O00499', 'O00629', 'O00712', 'O00716', 'O14786', 'Q9UJX3']

def test_load_protein_attribute(TS1_domains2_sites_tracks_protein_attributes):    

    assert TS1_domains2_sites_tracks_protein_attributes.protein('O00401').attribute('GO3') == 'tags from a separate'

    assert len(TS1_domains2_sites_tracks_protein_attributes.protein('O00401').attributes) == 5







"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""


def test_site_track_functions(TS1_domains2_sites_tracks_protein_attributes):
    # check we corretly count the number of unique domains

    # deep cut that relies on a bunch of things working
    assert TS1_domains2_sites_tracks_protein_attributes.protein('O00401').site(242)[1].attribute('ENZYME',safe=False) == 'TNK21'

    assert TS1_domains2_sites_tracks_protein_attributes.protein('O00401').site(307)[1].symbol == 'R'

    print(TS1_domains2_sites_tracks_protein_attributes.protein('O00401').sequence[1])
    assert TS1_domains2_sites_tracks_protein_attributes.protein('O00401').site(307)[1].symbol == TS1_domains2_sites_tracks_protein_attributes.protein('O00401').site(307)[1].residue

    


def test_site_domain_functions(TS1_domains2_sites_tracks_protein_attributes):

    c= 0
    for site in TS1_domains2_sites_tracks_protein_attributes.sites:

        d = site.get_domains()
        
        for domain in d:
            assert domain.start <= site.position
            assert domain.end >= site.position
            c= c+1


    # 126 is expected number for this file
    assert c == 126
                
        



"""
uniprot.py

From the SHEPHARD package
Sequence-based Hierachical and Extendable Platform for High-throughput Analysis of Region of Disorder
Ginell & Holehouse, 2020-2022

Handles all I/O associated with uniprot-derived files.

"""

from shephard.exceptions import UtilitiesException
import protfasta
from . import fasta

## ------------------------------------------------------------------------
##
def uniprot_accession_from_line(line):
    """
    Function that converts a header from a uniprot fasta file
    to extract the uniprot ID. This an example of the type of function
    that can be passed to quickstart using the extract_unique_id argument.

    This function assumes the uniprot-standard format for the header
    file has been maintained - i.e.

    >>> >xx|ACCESSION|xxxx

    where ACCESSION is the uniprot accession. 

    Parameters
    -----------

    line : string
        String where we expect the uniprot ID to be contained within two 'pipe' 
        characters ('|'). 

    Returns
    -----------
    string
        Returns the uniprot ID, although this is not formally validated. 
        However, assuming the string follows standard uniprot fasta header 
        conventions this should be true!

    """
    try:
        return line.split('|')[1].strip()
    except:
        raise UtilitiesException('Unable to parse string [%s] to identify uniprot ID' %(line))

        
        
## ------------------------------------------------------------------------
##
def uniprot_fasta_to_proteome(filename, 
                              proteome = None,
                              force_overwrite=False,
                              invalid_sequence_action='fail'):
                              
    """
    Stand alone function that allows the user to build a proteome from a 
    standard FASTA file downloaded from UniProt

    This function assumes the uniprot-standard format for the header
    file has been maintained - i.e.

    >>> >xx|ACCESSION|xxxx

    Where ACCESSION is the uniprot accession and will be used as the 
    unique_ID
    
    Parameters
    ------------

    filename : string
        Name of the FASTA file we're going to parse in. Note the protein 
        name will be defined as the full FASTA header for each entry.
        
    proteome : Proteome
        If a Proteome object is provided the FASTA file will be read and 
        added to the existing proteome, whereas if set to None a new 
        Proteome will be generated.

    force_overwrite : bool (default  = False)
        If this flag is set to true  and we encounter a unique_ID that is 
        already in the proteome the newer value overwrites the older one. 
        This is mostly useful if you are adding in a file with known 
        duplicate entries OR combining multiple FASTA files where you know 
        there's some duplications. Important - if we're building unique IDs
        based on numerical record indices then EVERY FASTA entry will be given 
        a unique_ID (meaning force_overwrite is irrelevant in this case).

    invalid_sequence_action : str (default = 'fail')
        Selector which defines the behaviour if a sequence with a non-
        standard amino acid is encountered. Valid options and their meaning
        are listed below:

            * ``ignore``  - invalid sequences are completely ignored

            * ``fail``    - invalid sequence cause parsing to fail and throw an exception
  
            * ``remove`` -  invalid sequences are removed

            * ``convert`` - invalid residues are converted to valid residues                            

            * ``convert-ignore`` - invalid sequences are converted to valid sequences and any remaining invalid residues are ignored.
    
    Returns 
    --------
    Proteome
        Returns an initialized Proteome object 
    
    """
    
    return fasta.fasta_to_proteome(filename, proteome=proteome, build_unique_ID=uniprot_accession_from_line, force_overwrite=force_overwrite, invalid_sequence_action=invalid_sequence_action)


## ------------------------------------------------------------------------
##
def uniprot_proteome_to_fasta(filename, proteome):                              
    """
    Stand alone function that allows the user to write a FASTA file from
    a Proteome under the assumption that the Proteome was built from a 
    uniprot FASTA.

    Practically, this just means that the Protein.name variable is used
    for the FASTA header, although the function will fail if duplicate
    headers are found.

    
    Parameters
    ------------

    filename : string
        Name of the FASTA file we're going to write sequences to

    proteome : Proteome
        The Proteome object from which FASTA file will be generated

    
    Returns 
    --------
    None
        No return variable but wll write to file 
    """

    out_dict = {}
    for p in proteome:

        header = p.name

        if header in out_dict:
            raise UtilitiesException(f'Duplicate name entries found in Proteome ({header}). Should not happen for UniProt headers')
        
        out_dict[header] = p.sequence

    protfasta.write_fasta(out_dict, filename, linelength=80)



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from . import exceptions
from shephard.exceptions import ShephardException

STANDARD_AAs = 'ACDEFGHIKLMNPQRSTVWY'


## ------------------------------------------------------------------------
##
def valid_keyword(keywordname, kw, kwlist):
    if kw not in kwlist:
        raise ShephardException("Provided keyword '%s' for argument %s but expected one of %s" % (kw, keywordname, str(kwlist)))


## ------------------------------------------------------------------------
##
def cast_or_none(value, cast_type):
    """
    Function that takes a value and a type (cast_type). If the value is None
    the fuction returns None. Otherwise the the cast_type will try and cast 
    the value. 
    
    Parameters
    ------------
    value : unknown
        Some kind of variable that can (hopefully) by 
    
    cast_type : some type of castable type
        A Python type that could be used to cast a variable, such as str, int,
        float etc.


    Returns
    -------
    
    None or casted type
    """

    if value is None:
        return None
    else:
    
        # envelope this in a try/except block to help the user out
        try:
            return cast_type(value)
        except Exception as e:
            raise ShephardException('In cast_or_none unable to cast passed variable, error below:\n%s' % str(e))
            


## ------------------------------------------------------------------------
##
def string_to_list_of_strings(inval):
    """
    Function that converts a string or list of
    strings to a list of strings. 

    """

    if type(inval) == str:
        return [inval]
    elif type(inval) == list:
        return inval

    else:
        raise exceptions.ShephardException('Expected either a single string or a list of strings')



## ------------------------------------------------------------------------
##
def numerical_average(l):
    """
    Simple function that returns the mean of a list or any iterable
    data type

    Parameters:
    ----------------
    l : list, np.ndarray, or any intertable container type that can be
        iterated over

    Returns:
    ----------------
    float
        Returns the average obtained by summing all elements and 
        dividing by number of elements


    """
    return numerical_sum(l)/len(l)


## ------------------------------------------------------------------------
##
def numerical_sum(l):
    """
    Simple function that returns the sum of a list or any iterable
    data type

    Parameters:
    ----------------
    l : list, np.ndarray, or any intertable container type that can be
        iterated over

    Returns:
    ----------------
    numerical
        Returns a value that reflects the sum of the elements, where type
        will be dictated by the Python rules for numerical addition .

    """


    t = 0

    for i in l:
        t = t + i

    return t

## ------------------------------------------------------------------------
##
def variable_is_dictionary(v, exc, msg, or_none=False):
    """
    General function to verify if a variable is a dictionary.

    Parameters
    -----------
    v : proposed dictionary
        Variable we are check if is a dictionary or not

    exc : Exception
        Subclass of exception that will be raised upon failured

    msg : str
        String to be passed to the exception when it's raised

    or_none : bool
        Variable that if is true returns True if 'v' is dict or None
        Otherwise only returns True if V is a dict

    Returns
    -------
    None or exception-rasied
        If the passed variable is a dictionary returns None, else
        raises an exception

    """

    if or_none:
        if v is None:
            return None

    if not isinstance(v, dict):
        raise exc(msg)

    return None






"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""
import re

## ------------------------------------------------------------------------
##
def build_mega_string(object_list, return_as_list=False): 
    
    """
    This takes a list of protein or domain SHEPHARD objects 
    and builds a single long str object that concatinates all
    object.sequence elements together (i.e. a "megastring").

    Allowed types for the object_list are Protein and Domain
    objects.

    This string can be used for simple statistical analysis
    of composition.

        
    Parameters
    ----------
    object_list : list 
        List of SHEPHARD objects with object.sequence variable,
        for example, a list of Domains or a list of Proteins

    return_as_list : bool
        If provided, rather than a single megastring, the 
        function returns a list of sequences from the objects
        in question.
        
    Returns
    ----------
    str or list
        Returns either a concatinated str object of the amino
        acid sequences associated with the passed object
    """


    megastring = ''
    for obj in object_list:  
        megastring = megastring + obj.sequence

    return megastring

## ------------------------------------------------------------------------
##
def find_string_positions(query, target, protein_indexing=True):
    """
    Returns list of start positions where 
    stringA is in stringB - including overlaps.

    Note that by default the indices use 1-indexing so that this works
    directly with protein sequence numbering. However, for manipulating
    Python strings this may be undesirable and 0 indexing may be better,
    in which case setting protein_indexing = False will address this.

    Practically, this uses the re regex expression under the hood 
    and searches left-to-right across the target, so if you want 
    to get fancier with your searching you can always pass in a 
    regular expression.

    Examples
    ----------------

    Conveninet regular expression syntax includes:

        1. ``'.'`` for wildcards (e.g. ``'L.P'`` would match an L and P around any other character

        2. ``[A|C]`` for requiring matching of a subset of residues (e.g. residue A and C).

    But the python re module has a fairly complex pattern matching
    ability

    Parameters
    --------------
    query : str
        The search query. 

    target : str
        The string that we'll search for 1 or more entries
        of the query

    protein_indexing : bool
        Flag which, if set to True, means the first residue in 
        a string indexes at '1' instead of '0' (as would be normal
        in Python. If set to False, then indexing is done from 0.
     

    Returns
    ----------
    list
         Returns a list with the start positions
    
    """
    if protein_indexing is True:
        offset = 1
    else:
        offset = 0

    return [s.start() + offset for s in re.finditer('(?=%s)' % (query), target)]



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (alex.holehouse@wustl.edu, g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from . import interface_tools 
import shephard.exceptions as shephard_exceptions
from shephard.exceptions import InterfaceException, ProteinException, SiteException

class _SitesInterface:

    def __init__(self, filename, delimiter='\t', skip_bad=True):
        """
        Expect files of the following format:
        
        A SHEPHARD sites file is a tab (or other) delineated file where each 
        line has the following convention::
    
               1        2          3       4      5   [      6            7        ...     n         ] 
            Unique_ID position site_type symbol value [key_1:value_1 key_2:value_2 ... key_n:value_n ]
    
        Each line has six required values and then can have as many key:value pairs as may be
        desired.

        Note that the first four arguments are required, while all of the 
        key:value pairs are optional. Key value must be separated by a ':', 
        but any delimiter (other than ':') is allowed. 

        Parameters
        ----------------
        
        filename : str
            Name of the shephard domains file to read

        delimiter : str (default = \t)
            String used as a delimiter on the input file. 

        skip_bad : bool (default = True)
            Flag that means if bad lines (lines that trigger an exception) 
            are encountered the code will just skip them. By default this is 
            true, which adds a certain robustness to file parsing, but could 
            also hide errors. Note that if lines are skipped a warning will be 
            printed (regardless of verbose flag). 


        """

        if delimiter == ':':
            raise InterfaceException('When parsing site file cannot use ":" as a delimeter because this is used to delimit key/value pairs (if provided)')


        with open(filename,'r') as fh:
            content = fh.readlines()

        ID2site={}
        
        linecount=0
        for line in content:

            linecount = linecount + 1

            # skip comment lines
            if interface_tools.is_comment_line(line):
                continue

            sline = line.strip().split(delimiter)

            try:
                unique_ID = sline[0].strip()
                position = int(sline[1].strip())
                site_type = sline[2].strip()
                symbol = sline[3].strip()

                # this enables the value to be None if you
                # write a symbol where there's no value associated
                # with a site
                tmp = sline[4].strip()
                if tmp == 'None':
                    value = None
                else:
                    value = float(tmp)

                attributes = {}
            except Exception as e:
                msg = f'Failed parsing file [{filename}] on line [{linecount}].\n\nException raised: {str(e)}\n\nline printed below:\n{line}'

                # should update this to also display the actual error...
                if skip_bad:
                    shephard_exceptions.print_warning(msg + "\nSkipping this line...")
                    continue
                else:
                    raise InterfaceException(msg)

            # if there's more parse attribute dictionary entries
            if len(sline) > 5:
                attributes = interface_tools.parse_key_value_pairs(sline[5:], filename, linecount, line)

            if unique_ID in ID2site:
                ID2site[unique_ID].append({'position':position, 'site_type':site_type, 'symbol':symbol, 'value':value, 'attributes':attributes})
            else:
                ID2site[unique_ID] =[{'position':position, 'site_type':site_type, 'symbol':symbol, 'value':value, 'attributes':attributes}]

        self.data = ID2site



##############################################
##                                          ##
##     PUBLIC FACING FUNCTIONS BELOW        ##
##                                          ##
##############################################


## ------------------------------------------------------------------------
##
def add_sites_from_file(proteome, filename, delimiter='\t', return_dictionary=False, safe=True, skip_bad=True, verbose=True):
    r"""
    Function that provides the user-facing interface for reading correctly 
    configured SHEPHARD sites files and adding those sites to the proteins 
    of interest.
    
    
    A SHEPHARD sites file is a tab (or other) delineated file where each 
    line has the following convention::
    
          1        2          3       4      5   [      6            7        ...     n         ] 
       Unique_ID position site_type symbol value [key_1:value_1 key_2:value_2 ... key_n:value_n ]
    
    Each line has six required values and then can have as many key:value pairs as may be
    desired.


    Parameters
    ----------
    proteome : Proteome
        Proteome object to which we're adding sites. Note that ONLY sites 
        for which a protein is found will be used. Protein-Site 
        cross-referencing is done using the protein's unique_ID which 
        should be the key used in the sites_dictionary

    filename : str
        Name of the shephard site file to be read

    delimiter : str (default = '\\t')
        String used as a delimiter on the input file. 

    return_dictionary : bool, default=False
        If set to true, this function will return the sites dictionary 
        and will NOT add that dictionary to the proteome - i.e. the 
        function basically becomes a parser for SHEPHARD-compliant        
        sites files. 

    safe : bool (default = True)
        If set to True then any exceptions raised during the site-adding 
        process (i.e. after file parsing) are acted on. If set to False, 
        exceptions simply mean the site in question is skipped. There are 
        various reasons site addition could fail (e.g. site falls outside 
        of protein position so if verbose=True then the cause of an exception 
        is also printed to screen. It is highly recommend that if you choose 
        to use safe=False you also set verbose=True. Default = True.
        
    skip_bad : bool (default = True)
        Flag that means if bad lines (lines that trigger an exception) are 
        encountered the code will just skip them. By default this is true, 
        which adds a certain robustness to file parsing, but could also hide 
        errors. Note that if lines are skipped a warning will be printed 
        (regardless of verbose flag). 

    verbose : bool (default = True)
        Flag that defines how 'loud' output is. Will warn about errors 
        on adding sites.
        
    Returns
    ---------
    None or dict
        If return_dictionary is set to False (default) then this function 
        has no return value, but the sites are added to the Proteome object 
        passed as the first argument. If return_dictionary is set to True 
        the function returns the parsed sites dictionary without adding the 
        newly-read sites to the proteome.

    """

    # check first argument is a proteome
    interface_tools.check_proteome(proteome, 'add_sites_from_file (si_sites)')

    sites_interface = _SitesInterface(filename, delimiter, skip_bad)

    if return_dictionary:
        return sites_interface.data


    # finally add the site from the dictionary generated by the SitesInterface parser
    add_sites_from_dictionary(proteome, sites_interface.data, safe, verbose)



## ------------------------------------------------------------------------
##
def add_sites_from_dictionary(proteome, sites_dictionary, safe=True, verbose=False):
    """
    Function that takes a correctly formatted Sites dictionary and will add 
    those Sites to the proteins in the Proteome.
    
    Sites dictionaries are key-value pairs, where the key is a unique_ID 
    associated with a given Protein, and the value is a list of dictionaries. 
    Each subdirectionay has the following elements::
    
        'position'   = site position
        'site_type'  = site type
        'symbol'     = site symbol 
        'value'      = site value 
        'attributes' = site attribute dictionary

    In this way, each site that maps to a give unique_ID will be added to 
    the associated protein. The use of a list of dictionaries (as opposed
    to a simple unique_ID:site_dictionary pairing) means multiple sites 
    for a single protein can be added at once.

    Parameters
    -------------

    proteome : Proteome
        Proteome object to which we're adding sites. Note that ONLY sites 
        for which a protein is found will be used. Protein:Site 
        cross-referencing is done using the protein's unique_ID        
        which should be the key used in the sites_dictionary

    sites_dictionary : dict
        A sites dictionary (defined above) is dictionary that maps a 
        unique_ID back to a list of dictionaries, where each 
        subdictionay has five elements, desribed above.

        Recall the only type-specific values (position and value) are 
        cast automatically when a site is added by the Protein object, 
        so there is no need to do that in this function too.

        Extra key-value paris in each sub-dictionary are ignored

    safe : bool (default = True)
        If set to True then any exceptions raised during the site-adding 
        process are acted on. If set to false, exceptions simply mean the 
        site in question is skipped. There are various reasons site addition 
        could fail (notably position of the site is outside of the protein 
        limits) and so if verbose=True then the cause of an exception is
        also printed to screen. It is highly recommend that if you choose to
        use safe=False you also set verbose=True

    verbose : bool (default = False)
        Flag that defines how 'loud' output is. Will warn about errors on 
        adding sites.

    Returns
    ---------
    None
        No return value, but adds all of the passed sites to the protein
    
    """
    
    for protein in proteome:
        if protein.unique_ID in sites_dictionary:
            for site in sites_dictionary[protein.unique_ID]:

                try:
                    position = site['position']
                    site_type = site['site_type']
                    symbol = site['symbol']
                    value = site['value']
                    try:
                        ad = site['attributes'] 
                    except:
                        ad = {}
                except Exception:
                    raise InterfaceException('When sites dictionary for key [%s] was unable to extract five distinct parametes. Entry is:\n%s\n'% (protein.unique_ID, site))

                # assuming we can read all five params try and add the site
                try:
                    protein.add_site(position, site_type, symbol, value, attributes = ad)                

                except ProteinException as e:
                    msg='- skipping site %s at %i on %s' %(site_type, position, protein)
                    if safe:
                        shephard_exceptions.print_and_raise_error(msg, e)
                    else:
                        if verbose:
                            shephard_exceptions.print_warning(msg)
                            continue
  

                  
## ------------------------------------------------------------------------
##
def write_sites(proteome, filename, delimiter='\t', site_types=None):
    r"""
    Function that writes out sites to file in a standardized format. Note 
    that attributes are converted to a string, which for simple attributes 
    is reasonable but is not really a viable stratergy for complex objects, 
    although this will not yeild and error.

    If a site_types list is provided, only site_types that match to
    strings in this list are written out.
    
    Parameters
    -----------
    proteome :  Proteome
        Proteome object from which the sites will be extracted from

    filename : str
        Filename that will be used to write the new sites file

    site_type : str (default = None)
        If provided, this is an identifier that allows you to specificy 
        a specific site type to write out.

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is the tab character ('\\t'), which is recommended to 
        maintain compliance with default SHEPHARD file-reading functions.     

    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        sites from this proteome written to disk.
        

    """

    with open(filename, 'w') as fh:
        for protein in proteome:
            for s in protein.sites:

                # if we're using site_types and the current sites 
                if site_types is not None:
                    if s.site_type not in site_types:
                        continue

                # build a line 
                # if the passed parameter site_types is being
                # used
                line = __build_site_line(s, delimiter)

                fh.write(f"{line}")



## ------------------------------------------------------------------------
##
def write_sites_from_list(site_list, filename, delimiter='\t'):
    r"""
    Function that writes out sites to a SHEPHARD sites file from a list
    of Site objects. 
    Note that attributes are converted to a string, which for simple 
    attributes is reasonable but is not really a viable stratergy for 
    complex objects, although this will not yeild and error.
            
    Parameters
    -----------

    site_list : List of Site objects
        List of site objects which will be written

    filename : str
        Filename that will be used to write the new sites file

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. Default is 
        '\\t' which is recommended to maintain compliance with default 
        `add_sites_from_file()` function
       
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        sites from this proteome written to disk.

    """

    # first check if items in the list are site objects
    for s in site_list:
        interface_tools.check_site(s, 'write_sites_from_list')

    with open(filename, 'w') as fh:

        # for each site in the list
        for s in site_list:

            # build a line 
            # if the passed parameter site_types is being
            # used
            line = __build_site_line(s, delimiter)

            fh.write(f"{line}")



## ------------------------------------------------------------------------
##
def __build_site_line(s, delimiter):
    """
    Internal function that takes a Site object and returns a line that can
    be written to a Sites file. This is called internally by functions that
    write Sites.

    Parameters
    ----------------------
    s : shephard.Site
        Site object being converted to a string

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is the tab character ('\\t'), which is recommended to 
        maintain compliance with default SHEPHARD file-reading functions.     

    Returns
    --------------
    str
        Returns a string that is ready to be written to file

    """

    # systematically construct each line in the file 
    line = ''
    line = line + str(s.protein.unique_ID) + delimiter
    line = line + str(s.position) + delimiter
    line = line + str(s.site_type) + delimiter                
    line = line + str(s.symbol) + delimiter

    # note last required element has no trailing delimiter
    line = line + str(s.value) 
    
    if s.attributes:
        for k in s.attributes:
            atrbt = interface_tools.full_clean_string(s.attribute(k))
            line = line + delimiter + f"{k}:{atrbt}"

    line = line + "\n"

    return line



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import numpy as np
import random
from . import site_tools
from shephard import general_utilities
from shephard import exceptions
from shephard.interfaces.interface_tools import check_protein

## ------------------------------------------------------------------------
##
def domain_overlap(domain_1, domain_2, check_origin=True):
    """
    Given two domains asks if their boundaries overlap. By default
    this expects the two domains in question to be from the same protein
    and checks this. If we dont want to enforce this assumption set
    check_origin to False.

    Parameters
    -----------
    domain_1 : shephard.domain.Domain
        The first domain object of interest

    domain_2 : shephard.domain.Domain
        The first domain object of interest

    check_origin : bool
        Flag that if set to True will cause an exception if domain_1 and
        domain_2 are from different proteins. If set to false, no such
        sanity checks are performed.

    Returns
    ----------
    boolean
        Returns true if the two domains overlap, else returns false
      
    """

    if check_origin:
        if domain_1.protein.unique_ID != domain_2.protein.unique_ID:
            raise exceptions.DomainException('Examining overlap of %s and %s but these are from different proteins' % (str(domain_1), str(domain_2)))
            
    return domain_overlap_by_position(domain_1.start, domain_1.end, domain_2.start, domain_2.end)



## ------------------------------------------------------------------------
##
def domain_overlap_fraction(domain_1, domain_2, check_origin=True):
    """
    Given two domains asks what fraction the shorter domain 
    overlaps the longer one with.
    Parameters
    -----------
    domain_1 : shephard.domain.Domain
        The first domain object of interest

    domain_2 : shephard.domain.Domain
        The first domain object of interest

    check_origin : bool (default = True)
        Flag that if set to True will cause an exception if domain_1 and
        domain_2 are from different proteins. If set to false, no such
        sanity checks are performed.

    Returns
    ----------
    float
        Returns a float between 0 and 1 that corresponds to what 
        fraction of the shorter domain overlaps with the longer domain.
      
    """

    if check_origin:
        if domain_1.protein.unique_ID != domain_2.protein.unique_ID:
            raise exceptions.DomainException('Examining overlap of %s and %s but these are from different proteins' % (str(domain_1), str(domain_2)))


    if len(domain_1) < len(domain_2):
        d_short = domain_1
        d_long = domain_2
    else:
        d_short = domain_2
        d_long = domain_1

    
    # ......OOOOOOOOOOOOOOOOOOOOOOO............       long
    #...XXXXXXXXXXXX?????                             short
    #
    if d_short.start < d_long.start:
        if d_short.end < d_long.start:
            return 0.0
        else:
            return ((d_short.end - d_long.start)+1)/len(d_short)

    # ......OOOOOOOOOOOOOOOOOOOOOOO.............     end
    #.                   ?????XXXXXXXXXXXX......     short
    #
    if d_short.end > d_long.end:
        if d_short.start > d_long.end:
            return 0.0
        else:

            # note - the +1 accounts for the fact we're using 
            # an inclusive counting 
            return ((d_long.end - d_short.start)+1)/len(d_short)

    # if we get here d_short.start equal to or larger than d_long start
    # and d_short end equal to or smaller than d_long end, so 100% overlap
    return 1.0
    


## ------------------------------------------------------------------------
##
def domain_overlap_by_position(boundary_start1, boundary_end1, boundary_start2, boundary_end2):
    """
    Given four sets of starting/ending positions, this function asks if 
    their boundaries overlap. 

    Parameters
    -----------
    boundary_start1 : int
        Position of domain 1 start

    boundary_end : int
        Position of domain 1 end

    boundary_start2 : int
        Position of domain 2 start

    boundary_end : int
        Position of domain 2 end

    Returns
    ----------
    boolean
        Returns true if the two domains overlap, else returns false
    """

    # note we do this swapping around which of the two domains passed as boundaries is 'a' and 'b' in 
    # the visual schematic below
    for x in [[boundary_start1, boundary_end1, boundary_start2, boundary_end2], [boundary_start2, boundary_end2,boundary_start1, boundary_end1]]:
        
        a_start = x[0]
        a_end   = x[1]
        b_start = x[2]
        b_end   = x[3]


        # scenario 1
        # AAAAAAAA
        #     BBBBBBBBB
        if a_start <= b_start and a_end >= b_start:
            return True

        # scenario 2
        #           AAAAAAAA
        #     BBBBBBBBB
        if a_start >= b_start and a_start <= b_end:
            return True

    return False
    



## ------------------------------------------------------------------------
##
def build_missing_domains(protein, new_domain_type = 'missing'):
    """
    Function which takes a protein and builds a set of domains that represent 
    the "empty spaces". Domains are returned as a list of domain dictionaries
    which can be added to a protein via the add_domains() function.

    This tool is stateless - i.e. it does not alter the passed protein
    but instead only generates a numerical list which could be added as 
    a track.

    One could always combine this directly with the add_domains() function
    into a single line - e.g.
    
    # this line will automatically add all the missing regions as 
    # 'missing' proteins to the domain
    protein.add_domains(build_missing_domains(protein))

    Parameters
    -----------
    protein : shephard.protein.Protein object
        Protein object over which sites are identified

    new_domain_type : str (default = 'missing')
        Name to assign to the 'empty' domains.

    Returns
    -------------
    list of domain dictionaries

        Returns a list of domain dictionaries which can be then parsed or
        added to a protein via the add_domains() function.
    
    """

    check_protein(protein, 'build_missing_domains()')

    ##
    ## NOTE -this function uses i0 indexing to keep things simple and then
    ## corrects at the end
    ##
    
    # first constrct an empty vector of 0s. We're going to build '1s' into
    # the positions in the sequence occupied by domains
    all_res = [0]*len(protein)
    
    # for each domain in the protein
    for d in protein.domains:

        # for each position in each domain (note this will overwrite
        # which is fine). NOTE that we cyle from d.start-1 to d.end because
        # start and end index from real-world positions, but range is a non
        # inclusive function (while the domain boundaries are inclusive). 
        for i in range(d.start-1, d.end):
            all_res[i] = 1
                
    # we are now left with an all_res list where the 0s represent regions
    # not assigned to any domain

    # if we start inside an empty domain set inside to true
    if all_res[0] == 0:
        start = 0
        inside=True

    # else set start to -1  and inside to false
    else:
        start = -1
        inside=False

    all_domains=[]

    # for each position in the sequence (starting at 1 because 
    # we already assessed 0 above. 
    for i in range(1,len(all_res)):

        # if we find ourselves in an empty space
        if all_res[i] == 0:
            if not inside:
                inside=True
                start = i
        # if we fine ourselves in a domain of some kind
        else:

            # if we were previously in an empty space
            if inside:
                inside = False

                # not here we're adding in i0 indexing
                all_domains.append([start, i-1])

    # if we ended and were inside when we ended then the last
    # empty domain spans to the final residue in the sequence
    if inside:
        all_domains.append([start,len(all_res)-1])

    new_domain_list =[]
    for d in all_domains:

        # note the +1 here is so we insert at residue positions
        # that correctly map to real-space positions (i.e.i1 indexing)

        # construct a domain dictionary 
        local_domain={}        
        local_domain['start'] = d[0]+1
        local_domain['end'] = d[1]+1
        local_domain['domain_type'] = new_domain_type

        new_domain_list.append(local_domain)

    return new_domain_list
            
        

## ------------------------------------------------------------------------
##
def build_domains_from_track_values(proteome, 
                                    track_name, 
                                    binerize_function, 
                                    domain_type, 
                                    gap_closure = 3, 
                                    minimum_region_size = 20, 
                                    extend_ends = None, 
                                    verbose = True):

    
    """
    Function which takes a Proteome and builds a set of domains based on
    values tracks in each Protein in that Proteome. This effectively 
    allows you to discretize some continous variable into distinct local 
    domains, which can often facilitate specific types of analysis. This
    conversion is done using a custom-passed binerize function which 
    converts a normal track into a track of 0s and 1s. Residues that
    are assigned a value of 1 will be included in a domain assuming they
    fall within a contigous region of sufficient size, as defined by
    the parameters gap_closure and minimum_region_size, as discussed 
    below.

    This function operates on an entire Proteome-level, and is stateless
    (i.e. does not directly alter the passed proteome). Instead, the 
    function dictionary where keys are unique_IDs of proteins and values 
    is a list of one or more Domain dictioinaries (with a start, end,
    and domain_type key:value pair).

    The domains dictionary can be added to a proteome using the 
    si_domains.add_domains_from_dictionary(). As an example, as possible 
    workflow is as follows:

    >>> d = build_domains_from_track_values(proteome, 'cool_track', trackfx)
    >>> si_domains.add_domains_from_dictionary(proteome, d)
    
    Under the hood, the function works by cycling through each protein, 
    extracting the track, and converting into domains.
        
    If a protein is too short or it lacks a given track, the protein is 
    skipped.
    
    Parameters
    -----------
    proteome : shephard.proteome.Proteome
        The Proteome which is going to be scanned for each track. Note that
        the underlying Proteome is not altered by this function

    track_name : string
        Name of the track to convert. If the track name does not exist in
        a given protein that protein is skipped. In this way, a Proteome
        where only a subset of Proteins have tracks can be parsed without
        issue. The track must be a values track - symbols tracks should be
        converted to a values track first to avoid issue.

    binerize_function : function
        A function which takes a track and converts it to 0 or 1 (binerize, 
        as in, make binary). This enables a complex and continous 
        track to be converted into a binary classification, which is 
        practically what a domain-assigment needs (yes/no inside domain).
        This function must take in a single variable (the track values) 
        and return a new list or numpy array that is the same length as 
        the track values but possesses only 0 and 1 in each element.

    domain_type : str
        String that defines the name of the new domains to create. Can in
        principle be anything.

    gap_closure : int (default = 3)
        Defines spacing between 1s or 0s that will be filled in to generate
        contigous stretches of 0s or 1s. This helps avoid a scenario where
        breaks in contigous stretches impede the definition of a domain 
        above a certain size, as defined by minimum_region_size. In general
        a value of 3 works reasonably well in most scenarios.

    minimum_region_size : int (default = 20)
        Defines the smallest size for a domain allowed. This can be varied
        depending on the question or data, and it may make sense to have 
        corresponding changes in gap_closure if this value becomes 
        substantially larger than a gap_closure of 3.

    extend_ends : int (default = None)
        This is a somewhat niche feature which, if set to a number, means 
        that we check the extend_ends-th value at the N- and C-terminus 
        of the binarized track, and if 1 set all values from that position
        to the N and/or C terminus to 1. This is provided because sometimes
        binerize functions will inherently struggle with the very ends of 
        sequences, so this provides a way to cast the first and last 
        extend_ends values to be 1. This is fairly specific and probably
        only worth using in a scenario where there is a clear issue

    verbose : bool (default = True)
        This flag enables the function to print statues every 500 proteins.
        If the binerize function is expensive this can be good to ensure 
        progress is proceeding.

    Returns
    -------------
    dict
        Returns a dictionary of key-value pairs, where each key is a unique
        ID and each value is a list of 1 or more domain dictionaries. This
        return dictionary can be directly added to a Proteome using the 
        Proteome.add_domains_from_dictonary() function.

    """
    
    new_domains = {}

    c = 0
    for protein in proteome:

        # if our protein is too short do not try and generate a domain
        if len(protein) < 3*gap_closure + 1:
            continue
        
        # this is the counter of proteins we've actually scanned - if
        # we hit a 500-protein milestone print status if verbose is true
        c = c + 1
        if verbose and c % 500 == 0:
            print('On %i of %i' %(c, len(proteome)))
        
        # safe = false so will return None if no track of that name
        # found, and if so we continue to next protein
        t = protein.track(track_name, safe=False)
        if t is None:
            continue

        # extract out the values
        t = t.values
        
        # and apply the binerize function to the values
        B = binerize_function(t)

        # add to help debugging
        if len(B) != len(protein):
            raise exceptions.DomainException('Binerize function failed to return an array or list that matches the protein sequence length')
            

        ## Part 1 - remove gapes
        for g in range(1, gap_closure+1):
            # first fill in 


            # for each position
            i = 0
            finished = False
            while not finished:

                p1 = i
                p2 = i + g
                p3 = i + 2*g
                p4 = i + 3*g

                # if the complete set of smaller regions ahead is empty or 
                # fully assigned skip ahead because nothing to do here...
                if np.sum(B[p1:p4]) == 0:
                    i = p4

                elif np.sum(B[p1:p4]) == 3*g:

                    # we jump to the p3 position (and NOT p4) as this allows us to skip along without
                    # discarding positions we need for filling. Note if we know everything is empty
                    # it doesn't matter and we can jump to p4
                    i = p3

                else:
                    # if we have gapsize number of hits
                    if np.sum(B[p1:p2])  == g:

                        # and if a gap away there is another gapsize 
                        if general_utilities.numerical_sum(B[p3:p4])  == g:                        
                            B[p2:p3] = [1]*g
                    
                    i = i + 1

                if i + 3*g >= len(B):
                    finished = True


        ## Part 2 - remove domains that are too small - we adde the '-' caps so we can use
        # replace and distinguish c/n terminal values
        B_string = '-'
        for i in B:
            if i == 1:
                B_string = B_string + "1"
            else:
                B_string = B_string + "0"

        B_string=B_string+'-'
        
        # for sizes of contigous stretches that are up minimum_region_size + 1
        # replace with empty ('0') strings
        for i in range(1, minimum_region_size + 1):      

            # 011110 -> 000000
            B_string = B_string.replace('0' + i*'1' + '0', '0' + i*'0' + '0')

            # -11110 -> -00000
            B_string = B_string.replace('-'+i*'1' + '0', '-'+i*'0' + '0')

            # 01111- -> 00000-
            B_string = B_string.replace('0' + i*'1'+'-' , '0'+ i*'0'+'-' )

            # -1111- -> -0000-
            B_string = B_string.replace('-' + i*'1'+'-' , '-'+ i*'0'+'-' )


        # 1 to -1 to cut off the artifical caps we added
        for i in range(1, len(B_string)-1):
            B[i-1] = int(B_string[i])

        ## Part 3 - extend ends if required
        if extend_ends:
                
            # note we don't check length of extend ends initially, so if they're too big for the sequence
            # dynamicaly resize them so they're 1/2 of the sequence length
            if extend_ends + 1 >= len(B):
                extend_ends_val = len(B)/2
            else:
                extend_ends_val = extend_ends
                
            if B[extend_ends_val + 1] == 1:
                B[0:extend_ends_val] = [1]*len(B[0:extend_ends_val])

            if B[:-(extend_ends_val + 1)] == 1:
                B[-extend_ends_val:] = [1]*len(B[0:extend_ends_val])
            
        ## Part 4 - extract domain boundaires
        local_domains=[]
        if B[0] == 1:
            inside = True
            start = 0
        else:
            inside = False
            
        # for each position
        for idx in range(0,len(B)):

            i = B[idx]
                
            if i == 1:
                if inside:
                    continue
                else:
                    inside = True
                    start = idx

            if i == 0:
                if inside:
                    inside = False
                    end = idx-1
                    local_domains.append({'start':start+1,'end':end+1,'domain_type':domain_type})

        # if we finished inside
        if inside:
            local_domains.append({'start':start+1,'end':len(B),'domain_type':domain_type})

        # if we found any local domains...
        if len(local_domains) > 0:
            new_domains[protein.unique_ID] = local_domains

    return new_domains


   
# used for test example - kept in case we want to expand or update this in          

# def calculate_domain_length(domain):
#    """
#    Function that returns a domain's length
#    
#    Parameters
#    -------------
#    domain : Domain
#    Domain object in question
#
#    Returns
#    ------------
#    int
#    Returns the length of the domain
#    """
#    
#    return len(domain)


                
            





import pytest

import shephard
from shephard.apis import uniprot  
from shephard import interfaces
import numpy as np
from shephard.exceptions import ProteinException 
import os



def test_remove_track():

    ##
    ## This code block just creates a proteome where we have 2 different
    ## typs of tracks
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)


    assert len(P.unique_track_names) == 0
    interfaces.si_tracks.add_tracks_from_file(P, pscore_track_file, mode='values')

    assert len(P.unique_track_names) == 1
    
    for protein in P:
        t = protein.track('pscore')
        protein.remove_track(t)
        break

    assert len(P.unique_track_names) == 1

    for protein in P:
        t = protein.track('pscore', safe=False)
        if t is not None:
            protein.remove_track(t, safe=False)
        
    assert len(P.unique_track_names) == 0


def test_remove_site():
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    assert len(P.unique_site_types) == 0

    interfaces.si_sites.add_sites_from_file(P, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    assert len(P.unique_site_types) == 4

    for protein in P:
        for site in protein.sites:
            protein.remove_site(site)

    assert len(P.unique_site_types) == 0
    interfaces.si_sites.add_sites_from_file(P, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))
    assert len(P.unique_site_types) == 4


    for protein in P:
        for site in protein.sites:
            if site.site_type == 'Sumoylation':
                protein.remove_site(site)

                # this should raise an exception
                with pytest.raises(ProteinException):
                    protein.remove_site(site)
                protein.remove_site(site, safe=False)


                # this should raise an exception
                with pytest.raises(ProteinException):
                    protein.remove_site('THIS IS NOT VALID')
                protein.remove_site('STILL NOT VALID', safe=False)
                


    assert len(P.unique_site_types) == 3            

    for protein in P:
        for site in protein.sites:
            if site.site_type == 'Sumoylation':
                protein.remove_site(site)


def test_remove_domains():
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    assert len(P.unique_domain_types) == 0

    interfaces.si_domains.add_domains_from_file(P, domain_file)
    assert len(P.unique_domain_types) == 1


    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr_and_others.tsv')
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_domains.add_domains_from_file(P, domain_file)
    assert len(P.unique_domain_types) == 2

    
    protein = P.protein('O00401')
    protein.add_domain(1,40,'TEST')
    assert len(P.unique_domain_types) == 3
    domain_obj = protein.domains[1]
    protein.remove_domain(domain_obj)
    assert len(P.unique_domain_types) == 2



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (alex.holehouse@wustl.edu, g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from shephard.exceptions import InterfaceException
from . import interface_tools 
from shephard.exceptions import ProteinException
import shephard.exceptions as shephard_exceptions

class _ProteinAttributesInterface:

    """
    Class whose sole purpose is to encapsulate and then store
    parsed Protein Attribute files. This is a hidden class and is not 
    accessible outside of this file
    
    """

    def __init__(self, filename, delimiter='\t', skip_bad=True):
        r"""
        Expect files of the followin format:

        Unique_ID, key1:value1, key2:value2, ..., keyn:valuen


        Parameters
        ----------------
        
        filename : str
            Name of the shephard domains file to read


        Other Parameters
        ----------------

        delimiter : str (default = '\\t')
            String used as a delimiter on the input file. 

        skip_bad : bool (default = True)
            Flag that means if bad lines (lines that trigger an exception) 
            are encountered the code will just skip them. By default this is 
            true, which adds a certain robustness to file parsing, but could 
            also hide errors. Note that if lines are skipped a warning will be 
            printed (regardless of verbose flag). 
            Default = True


        """

        if delimiter == ':':
            raise InterfaceException('When parsing domain file cannot use ":" as a delimeter because this is used to delimit key/value pairs (if provided)')

        with open(filename,'r') as fh:
            content = fh.readlines()

        ID2ADs={}

        linecount=0
        for line in content:

            linecount = linecount + 1

            # skip comment lines
            if interface_tools.is_comment_line(line):
                continue

            sline = line.strip().split(delimiter)

            # try
            try:
                unique_ID = sline[0].strip()
                attributes = {}                
            except Exception as e:

                msg = 'Failed parsing file [%s] on line [%i].\n\nException raised: %s\n\nline printed below:\n%s'%(filename, linecount, str(e), line)

                # should update this to also display the actual error...
                if skip_bad:
                    shephard_exceptions.print_warning(msg + "\nSkipping this line...")
                    continue
                else:
                    raise InterfaceException(msg)

            # if some key/value pairs were included then parse these out one at a time
            if len(sline) > 1:
                attributes = interface_tools.parse_key_value_pairs(sline[1:], filename, linecount, line)
            else:
                # skip over empty entries
                continue
  
            if unique_ID in ID2ADs:
                ID2ADs[unique_ID].append(attributes)
            else:
                ID2ADs[unique_ID] = [attributes]

        self.data = ID2ADs



##############################################
##                                          ##
##     PUBLIC FACING FUNCTIONS BELOW        ##
##                                          ##
##############################################


## ------------------------------------------------------------------------
##
def add_protein_attributes_from_dictionary(proteome, protein_attribute_dictionary, safe=True, verbose=True):
    r"""
    Function that takes a correctly formatted protein_atttribute dictionary
    and will add those attributes to the proteins in the Proteome.
    
    protein attribute dictionaries are key-value pairs, where the key is a 
    unique ID and the value is a list of dictionaries. For each sub-dictionary, 
    the key-value pair reflects the attribute key-value pairing.

    Parameters
    ----------
    proteome : Proteome Object
        Proteome object to which attributes will be added

    protein_attribute_dictionary : dict
        Dictionary that defines protein attributes. This is slightly 
        confusing, but the keys for this dictionary is a unique 
        protein IDs and the values is a list of dictionaries. Each of 
        THOSE sub-dictionaries has one (or more) key:value pairs that 
        define key:value pairs that will be associated with the protein 
        of interest.

    safe : boolean (default = True)
        If set to True then any exceptions raised during the process of 
        adding a protein_attribute are further raised. If set to False, 
        exceptions simply mean the protein_attribute in question is skipped.         
        Note if set to False, pre-existing protein_attributes with the same 
        name would be silently overwritten (although this is not consider an 
        error), while overwriting will trigger an exception if safe=True.
        Default = True
        
        The only reason protein attribute addition could fail is if the 
        attribute already exists, so this is effectively a flag to define 
        if pre-existing attributes should be overwritten (False) or not 
        (True).
    
    verbose : bool (default = True)
        Flag that defines how 'loud' output is. Will warn about errors on 
        adding attributes.

    Returns
    -----------
    None
        No return value, but attributes are added to proteins in the Proteome 
        object passed as the first argument
    """

    # check first argument is a Proteome
    interface_tools.check_proteome(proteome, 'add_protein_attributes (si_protein_attributes)')
    
    for protein in proteome:
        if protein.unique_ID in protein_attribute_dictionary:

            # note here each AD is its own dictionary
            for AD in protein_attribute_dictionary[protein.unique_ID]:

                # for each attribute-key
                for k in AD:            

                    # get the value
                    v = AD[k]

                    try:
                        protein.add_attribute(k, v, safe=safe)
                    except ProteinException as e:
                        msg='- skipping attribute entry on protein %s (key: %s) ' % (protein.unique_ID, k)
                        if safe:
                            shephard_exceptions.print_and_raise_error(msg, e)
                        else:
                            if verbose:
                                shephard_exceptions.print_warning(msg)
                                continue


## ------------------------------------------------------------------------
##
def add_protein_attributes_from_file(proteome, 
                                     filename, 
                                     delimiter='\t', 
                                     return_dictionary=False, 
                                     safe=True, 
                                     skip_bad=True, 
                                     verbose=True):
    r"""
    Function that takes a correctly formatted 'protein attributes' file and 
    reads all attributes into the proteins in the passed proteome.

    The function expects protein attribute files to have the following 
    format:

    One protein defined per line (although the same protein can appear 
    multiple times)

    >>> Unique_ID, key1:value1, key2:value2, ..., keyn:valuen

    A couple of key points here:

    - The default delimiter is tabs ('\\t') but this can be changed with 
      the delimiter argument

    - Key value must be separated by a ':', as a result any delimiter 
      (other than ':') can be used, but ':' is reserved for this role
      
    Parameters
    ----------
    proteome : Proteome Object
        Proteome object to which attributes will be added.

    filename : str
        Name of the shephard protein attributes file to read.

    delimiter : str (default = '\\t')
        String used as a delimiter on the input file. 

    return_dictionary : bool (default = False)
        If set to True, this function will return the protein_attributes 
        dictionary and will NOT add that dictionary to the proteome - 
        i.e. the function basically becomes a parser for SHEPHARD-compliant        
        protein_attributes files. 

    safe : bool (default = True)
        If set to True then any exceptions raised during the 
        protein_attribute-adding process are acted on. If set to False, 
        exceptions simply mean the protein_attribute in question is skipped.         
        Note if set to False, pre-existing protein_attributes with the same 
        name would be silently overwritten (although this is not consider an 
        error), while overwriting will trigger an exception if safe=True.
        
        The only reason protein attribute addition could fail is if the 
        attribute already exists, so this is effectively a flag to define 
        if pre-existing attributes should be overwritten (False) or not 
        (True).

    skip_bad : bool (default = True)
        Flag that means if bad lines (lines that trigger an exception) are 
        encountered the code will just skip them. By default this is true, 
        which adds a certain robustness to file parsing, but could also hide 
        errors. Note that if lines are skipped a warning will be printed 
        (regardless of verbose flag). skip_bad exclusively influences the 
        file-reading part of the process.
        
    verbose : bool (default = True)
        Flag that defines how 'loud' output is. Will warn about errors on 
        adding attributes.

    Returns
    ----------- 
    None or dict
        If return_dictionary is set to False (default) then this function 
        has no return value, but the protein_attributes are added to the 
        Proteome object passed as the first argument. If return_dictionary
        is set to True the function returns the parsed domains_dictionary 
        without adding the newly-read protein_attributes to the proteome.

    """

    # check first argument is a proteome
    interface_tools.check_proteome(proteome, 'add_attributes_from_file (si_protein_attributes)')

    # next read in the file
    protein_attribute_interface = _ProteinAttributesInterface(filename, 
                                                              delimiter, 
                                                              skip_bad=skip_bad)

    if return_dictionary:
        return protein_attribute_interface.data


    # finally add the domains from the dictionary generated by the DomainsInterface parser
    add_protein_attributes_from_dictionary(proteome, 
                                           protein_attribute_interface.data, 
                                           safe=safe, 
                                           verbose=verbose)



## ------------------------------------------------------------------------
##
def write_protein_attributes(proteome, filename, delimiter='\t'):
    r"""
    Function that writes out protein attributes to file in a standardized 
    format. Note that attributes are converted to a string, which for simple 
    attributes is reasonable but is not really a viable stratergy for 
    complex objects, although this will not yeild and error.
    
    Parameters
    -----------
    proteome :  Proteome object
        Proteome object from which the domains will be extracted from

    filename : str
        Filename that will be used to write the new domains file

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is '\t', which is recommended to maintain compliance 
        with default `add_protein_attributes_from_file()` function.        
        
    Returns
    --------
    None
        No return type, but generates a new file with the complete set 
        of protein attributes from this proteome written to disk.
        
    """

    with open(filename, 'w') as fh:
        for protein in proteome:
            if len(protein.attributes) > 0:

                line = protein.unique_ID

                for k in protein.attributes:

                    atrbt = interface_tools.full_clean_string(protein.attribute(k))

                    line = line + delimiter +  "%s:%s" %(k, atrbt)

                line = line + "\n"

                fh.write(line)



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""
import os


# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions


def get_version():
    return __version__

_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    """
    This function 
    """
    return os.path.join(_ROOT, 'data', path)

from .proteome import Proteome



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (alex.holehouse@wustl.edu, g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from . import si_sites
from . import si_domains
from . import si_tracks
from . import si_protein_attributes
from . import si_proteins 



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import pytest
from shephard.exceptions import ProteinException, DomainException
import numpy as np
import shephard
from shephard.apis import uniprot
from shephard import interfaces
from shephard.tools import track_tools

names_list = ['O00401', 'O00470', 'O00472', 'O00499', 'O00629', 'O00712', 'O00716', 'O14786', 'Q9UJX3']


def test_domain_to_track_1(TS1_domains2_sites_tracks):    
    """
    Note if any of these succeed the test suite will fail because it incrases
    """



    # build tracks from domains
    track_dict = track_tools.build_track_from_domains(TS1_domains2_sites_tracks)

    # check they're the right length
    for k in track_dict:
        assert len(track_dict[k]) == len(TS1_domains2_sites_tracks.protein(k))

    # check positions inside the domain are in fact set to 1
    for k in track_dict:
        for domain in TS1_domains2_sites_tracks.protein(k).domains:
            assert track_dict[k][domain.start-1] == "1"
            assert track_dict[k][domain.end-1] == "1"
            assert track_dict[k][domain.end-len(domain)] == "1"



    


def test_domain_to_track_2(TS1_domains2_sites_tracks):    
    """

    """



    # build tracks from domains
    track_dict = track_tools.build_track_from_domains(TS1_domains2_sites_tracks, domain_type='pscore_domain')


    # check they're the right length
    for k in track_dict:
        assert len(track_dict[k]) == len(TS1_domains2_sites_tracks.protein(k))

    # check positions inside the domain are in fact set to 1
    for protein in TS1_domains2_sites_tracks:
        k = protein.unique_ID

        for domain in protein.domains:

            if domain.domain_type == 'pscore_domain':
                print(domain.protein.unique_ID)
                assert track_dict[k][domain.start-1] == "1"
                assert track_dict[k][domain.end-1] == "1"
                assert track_dict[k][domain.end-len(domain)] == "1"



    # has no pscore_domains but does have an IDR domain
    k = 'O00629'
    for domain in TS1_domains2_sites_tracks.protein(k).domains:
        if domain.domain_type == 'IDR':
            
            # should NOT be a 1 because these domains are not pscore domains
            assert track_dict[k][domain.start-1] == "0"
            assert track_dict[k][domain.end-1] == "0"
            assert track_dict[k][domain.end-len(domain)] == "0"







import pytest

import shephard
from shephard.apis import uniprot  
from shephard.interfaces import si_domains
import numpy as np
from shephard.exceptions import ProteinException, DomainException
import random


test_data_dir = shephard.get_data('test_data')

# ....................................................................................................
#
#
def test_add_domains_file():


    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file)
    
    # this should fail because already added
    with pytest.raises(ProteinException):
        si_domains.add_domains_from_file(P, domain_file)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file, autoname=True)

    print('')
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file, autoname=False)

    # autoname allows 2 apparetly identical domain files to be added
    si_domains.add_domains_from_file(P, domain_file, autoname=True)

    # autoname allows 2 apparetly identical domain files to be added
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file, autoname=False, skip_bad=True)

def test_add_domain_attribute():


    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file)
    prot = P.protein('O00401')
    domain = prot.domains[0]
    domain.add_attribute('test_attribute', 1)

    assert domain.attribute('test_attribute') == 1

    # this should fail
    with pytest.raises(DomainException):
        domain.add_attribute('test_attribute', 20)

    # because the operation above should have failed, this too should
    # have failed
    assert domain.attribute('test_attribute') == 1

    domain.add_attribute('test_attribute', 20, safe=False)
    assert domain.attribute('test_attribute') == 20

    assert len(domain.attributes) == 1
    domain.add_attribute('another_test_attribute', 'testval')
    assert len(domain.attributes) == 2


    with pytest.raises(DomainException):
        assert domain.attribute('does not exist') == 20

    # check this returns none
    assert domain.attribute('does not exist', safe=False) is None


    
def test_write_domain_with_attributes():

    # this setup was also tested in test_add_domain_attribute
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    si_domains.add_domains_from_file(P, domain_file)
    prot = P.protein('O00401')
    domain = prot.domains[0]
    domain.add_attribute('test_attribute_1', 1)
    domain.add_attribute('test_attribute_cat', 'cat')


def test_write_domains():

    
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    n_domains = 3

    uid2domain_info = {}
    for p in TS1:

        
        uid2domain_info[p.unique_ID] = []
        for idx in range(n_domains):
            s = random.randint(1,len(p))
            e = random.randint(1,len(p))
            if s > e:
                start = e
                end = s
            else:
                start = s
                end = e
                
            if start  == 0:
                start = start +1

            if end > len(p):
                end = len(p)-1
            

            uid2domain_info[p.unique_ID].append([start,end,idx])

            p.add_domain(start, end, f'test_domain_{idx}', attributes={'att1':'test'})

    si_domains.write_domains(TS1, 'output_test/test_domains.tsv')


    TS2 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))
    si_domains.add_domains_from_file(TS2, 'output_test/test_domains.tsv')

    # check domain positions can be read in correctly
    for p in TS2:

        assert p.get_domains_by_type('test_domain_0')[0].start == uid2domain_info[p.unique_ID][0][0]
        assert p.get_domains_by_type('test_domain_1')[0].start == uid2domain_info[p.unique_ID][1][0]
        assert p.get_domains_by_type('test_domain_2')[0].start == uid2domain_info[p.unique_ID][2][0]


        assert p.get_domains_by_type('test_domain_0')[0].end == uid2domain_info[p.unique_ID][0][1]
        assert p.get_domains_by_type('test_domain_1')[0].end == uid2domain_info[p.unique_ID][1][1]
        assert p.get_domains_by_type('test_domain_2')[0].end == uid2domain_info[p.unique_ID][2][1]



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import pytest
import shephard
from shephard import proteome
from shephard.apis import uniprot, fasta
from shephard.exceptions import ProteomeException

names_list = ['O00401', 'O00470', 'O00472', 'O00499', 'O00629', 'O00712', 'O00716', 'O14786', 'Q9UJX3']

def test_proteome_len(TS1):    
    assert len(TS1) == 9
    assert len(TS1) == len(TS1.proteins)

    
def test_proteins_retrieve(TS1):
    # checks that all the unique IDs are there as expected
    for protein in TS1.proteins:
        assert protein in names_list


def test_record_and_unique_ID_match(TS1):
    # checks when we iterate we get a protein
    for p in TS1.proteins:
        assert p == TS1.protein(p).unique_ID



def test_iterator(TS1):
    for p in TS1:
        assert str(type(p)) == "<class 'shephard.protein.Protein'>"



def test_unique_domain_types(TS1, TS1_domains, TS1_domains2, TS1_domains2_sites, TS1_domains2_sites_tracks):
    # check we corretly count the number of unique domains

    assert len(TS1.unique_domain_types) == 0
    assert len(TS1_domains.unique_domain_types) == 1
    assert len(TS1_domains2.unique_domain_types) == 2
    assert len(TS1_domains2_sites.unique_domain_types) == 2
    assert len(TS1_domains2_sites_tracks.unique_domain_types) == 2



def test_unique_site_types(TS1, TS1_domains, TS1_domains2, TS1_domains2_sites, TS1_domains2_sites_tracks):
    # check we corretly count the number of unique sites

    assert len(TS1.unique_site_types) == 0
    assert len(TS1_domains.unique_site_types) == 0
    assert len(TS1_domains2.unique_site_types) == 0
    assert len(TS1_domains2_sites.unique_site_types) == 15
    assert len(TS1_domains2_sites_tracks.unique_site_types) == 15



def test_domain_retriev(TS1_domains):
    # check we corretly can get all domains
    assert len(TS1_domains.domains) == 21


def test_domain_retriev(TS1_domains2_sites):
    # check we corretly can get all sites
    assert len(TS1_domains2_sites.sites) == 209

def test_protein(TS1_domains2_sites):
    # check we corretly access a valid protein

    # get sequence
    len(TS1_domains2_sites.protein('O00629')) == 521
    len(TS1_domains2_sites.protein('O00629').domains) == 2
    len(TS1_domains2_sites.protein('O00629').sites) == 16

def test_add_protein():

    # creating proteome and adding protein
    test_data_dir = shephard.get_data('test_data')
    P = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'))
    assert len(P.protein('O00401')) == 505
    assert len(P.protein('O00470')) == 390
    assert len(P.protein('O00472')) == 640
    assert len(P.protein('O00499')) == 593
    assert len(P.protein('O00629')) == 521
    assert len(P.protein('O00712')) == 420
    assert len(P.protein('O00716')) == 465
    assert len(P.protein('O14786')) == 923
    assert len(P.protein('Q9UJX3')) == 599

    # creating a proteome from a FASTA file (using defaul unique key)
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'))
    assert len(P.protein('1')) == 390
    assert len(P.protein('2')) == 640
    assert len(P.protein('3')) == 593
    assert len(P.protein('4')) == 521
    assert len(P.protein('5')) == 420
    assert len(P.protein('6')) == 465
    assert len(P.protein('7')) == 923
    assert len(P.protein('8')) == 599

    # create a proteome where FASTA header is used as uniqueID 
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), use_header_as_unique_ID=True)
    assert len(P.protein('sp|O00470|MEIS1_HUMAN Homeobox protein Meis1 OS=Homo sapiens OX=9606 GN=MEIS1 PE=1 SV=1')) == 390
    assert len(P.protein('sp|O00472|ELL2_HUMAN RNA polymerase II elongation factor ELL2 OS=Homo sapiens OX=9606 GN=ELL2 PE=1 SV=2')) == 640
    assert len(P.protein('sp|O00499|BIN1_HUMAN Myc box-dependent-interacting protein 1 OS=Homo sapiens OX=9606 GN=BIN1 PE=1 SV=1')) == 593
    assert len(P.protein('sp|O00629|IMA3_HUMAN Importin subunit alpha-3 OS=Homo sapiens OX=9606 GN=KPNA4 PE=1 SV=1')) == 521
    assert len(P.protein('sp|O00712|NFIB_HUMAN Nuclear factor 1 B-type OS=Homo sapiens OX=9606 GN=NFIB PE=1 SV=2')) == 420
    assert len(P.protein('sp|O00716|E2F3_HUMAN Transcription factor E2F3 OS=Homo sapiens OX=9606 GN=E2F3 PE=1 SV=1')) == 465
    assert len(P.protein('sp|O14786|NRP1_HUMAN Neuropilin-1 OS=Homo sapiens OX=9606 GN=NRP1 PE=1 SV=3')) == 923
    assert len(P.protein('sp|Q9UJX3|APC7_HUMAN Anaphase-promoting complex subunit 7 OS=Homo sapiens OX=9606 GN=ANAPC7 PE=1 SV=4')) == 599

    

    
    # check manually adding proteomes    
    local_seq = 'PPPPP'
    P.add_protein(local_seq, '5pp', 'U5P')
    assert P.protein('U5P').sequence == local_seq
    assert P.protein('U5P').name == '5pp'

    # should trigger exception
    with pytest.raises(ProteomeException):
        P.add_protein(local_seq, '5pp', 'U5P')
    P.add_protein('ASDF', '5pp', 'U5P', force_overwrite=True)
    assert P.protein('U5P').sequence == 'ASDF'

    protein_list = []
    p1 = {'sequence': 'ASDFGH', 'name': "Test protein 1", 'unique_ID':1.23, "attributes":None}
    protein_list.append(p1)
    

    # check this works
    P = proteome.Proteome(protein_list)
    print(P.proteins)
    assert P.protein(1.23).sequence == 'ASDFGH'
    assert P.protein("1.23").sequence == 'ASDFGH'

    P.remove_protein(1.23)
    with pytest.raises(ProteomeException):
        assert P.protein(1.23).sequence == 'ASDFGH'
        

    
            


#test_add_protein()






import shephard
from shephard.interfaces import si_uniprot, si_domains, si_tracks, si_sites
from shephard import quickstart
import pytest
import sys


TS1_FILE=['testset_1.fasta', 'TS1_domains_idr.tsv', 'TS1_domains_pscore.tsv', 'TS1_sites.tsv', 'TS1_tracks_pscore.tsv']
test_data_dir = shephard.get_data('test_data')

def build_proteome(fn):
    return quickstart.quickstart('%s/%s' % (test_data_dir,fn),  extract_unique_ID=si_uniprot.extract_unique_ID_uniprot)


p = build_proteome(TS1_FILE[0])
si_domains.read_in_domains(p, '%s/%s' %(test_data_dir, TS1_FILE[1]))
si_domains.read_in_domains(p, '%s/%s' %(test_data_dir, TS1_FILE[2]))
si_sites.read_in_sites(p, '%s/%s' %(test_data_dir, TS1_FILE[3]))
si_tracks.read_in_tracks(p, '%s/%s' %(test_data_dir, TS1_FILE[4]))



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from . import general_utilities
from . import sequence_utilities
from .exceptions import ProteinException, SiteException


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# Class that defines a Site in sequence
#
class Site:
    
    def __init__(self, position, site_type, protein, symbol=None, value=None, attributes=None):

        """
        Sites are defined sub-positions within a protein that map to a 
        specific residue.
    
        Proteins contain a list of 0 or more sites, and each site is 
        associated with the protein it originates from via the linking 
        protein object.


        Parameters
        -------------

        position : int
            Position in sequence associated with this site

        site_type :  str
            Identifier for the site type

        protein : Protein
            Protein object for which this site is part of
        
        symbol : str (default = None)
            Symbol associated with the site - a string-based representation 
            of something specific to the site. For a mutation that could be 
            the residue the native residue mutates to, for example. 

        value : float (default = None)
            Value associated with the site - a numerical value (cast to a 
            float). 
        
        attributes : dict (default = None)
            The attributes dictionary provides a key-value pairing for 
            arbitrary information. This could include different types of 
            identifies, track generator functions, a set of Site partners, 
            or anything else one might wish to associated with the track as 
            a whole. 
        
        """


        # absolute position in protein associated with the site
        self._position  = int(position)
        
        # reference to the protein object from which the site was taken
        self._protein   = protein

        # the site type identifier
        self._site_type = str(site_type)

        # a numerical value associated with the site
        self._value     = general_utilities.cast_or_none(value, float)

        # a symbol associated with the site
        self._symbol    = general_utilities.cast_or_none(symbol, str)

        # verify that the attributes dictionary is a dictionary
        general_utilities.variable_is_dictionary(attributes, SiteException, 'attributes argument passed to site %i in protein %s is not a dictionary' %(self._position, self._protein), or_none=True)
        
        if attributes is None:
            self._attributes = {}
        else:
            self._attributes = attributes

        # update the proteome if this is a novel type of site
        protein.proteome.__update_site_types(self._site_type)


    ## ------------------------------------------------------------------------
    ##
    @property
    def residue(self):
        """
        Returns the amino acid residue associated with the site position as
        a string.
        """
        return self._protein.residue(self._position)


    ## ------------------------------------------------------------------------
    ##
    @property
    def position(self):
        """
        Returns the actual sequence indexed position as an int (recall protein
        indexing starts at 1).
        """
        return self._position


    ## ------------------------------------------------------------------------
    ##
    @property
    def protein(self):
        """
        Return the Protein object this site is found within
        """
        return self._protein


    ## ------------------------------------------------------------------------
    ##
    @property
    def site_type(self):
        """
        Returns the site type (string)
        """
        return self._site_type


    ## ------------------------------------------------------------------------
    ##
    @property
    def symbol(self):
        """
        Returns the symbol associated with this site. Note a symbol is either
        None or a str type.
        """
        return self._symbol


    ## ------------------------------------------------------------------------
    ##
    @property
    def value(self):
        """
        Returns the value associated with this site. Note a value is either
        None or a float type.
        """
        return self._value

    ## ------------------------------------------------------------------------
    ##
    def update_site_value(self, new_value):
        """
        Function that updates the site value. Values must be
        numerical (float) or None.
        
        Parameters
        -----------
        new_value : float (or None)
            Updated value

        Returns
        -----------
        None
            Nothing but sets the value to be the new value

        """
        if new_value is None:
            self._value = None
        else:
            self._value = float(new_value)

    ## ------------------------------------------------------------------------
    ##
    def update_site_symbol(self, new_symbol):
        """
        Function that updates the site_symbol. The site tyoe must be
        a string. Note this function also updates the proteome list of
        non-redudant sites
        
        Parameters
        -----------
        new_symbol : str (or None)
            Updated symbol

        Returns
        -----------
        None
            Nothing but sets the symbol to be the new symbol

        """
        if new_symbol is None:
            self._symbol = None
        else:
            self._symbol = str(new_symbol)





     ###################################
    ##                               ##
    ##     ATTRIBUTE FUNCTIONS       ##
    ##                               ##
    ###################################


    ## ------------------------------------------------------------------------
    ##
    @property
    def attributes(self):
        """
        **[Property]**: Provides a list of the keys associated with every 
        attribute associated with this Site.        

        Returns
        -------
        list
            returns a list of the attribute keys associated with the protein. 
        """

        return list(self._attributes.keys())



    ## ------------------------------------------------------------------------
    ##
    def attribute(self, name, safe=True):

        """
        Function that returns a specific attribute as defined by the name. 

        Recall that attributes are name : value pairs, where the 'value' 
        can be anything and is user defined. This function will return 
        the value associated with a given name.
        
        

        Parameters
        ----------------
        name : str
             The attribute name. A list of valid names can be found by 
             calling the ``<Site>.attributes()`` (which returns a list 
             of the valid names)
             
        safe : bool (default = True)
            Flag which if true with throw an exception if an attribute 
            with the same name already exists.
                        
        Returns
        ---------
        Unknown 
            Will either return whatever was associated with that attribute 
            (which could be anything) or None if that attribute is missing.
                    
        """

        # if name is in the _atributes dictionary the  return
        if name in self._attributes:
            return self._attributes[name]
        else:

            # else if safe was passed raise an exception if that attribute was missing
            if safe:
                raise SiteException('Requesting attribute [%s] from Site [%s] but this attribute has not been assigned' % (name, str(self))) 

            # if safe not passed just return None
            else:
                return None
                


    ## ------------------------------------------------------------------------
    ##
    def add_attribute(self, name, val, safe=True):
        """
        Function that adds an attribute. Note that if safe is true, this 
        function will raise an exception if the attribute is already present. 
        If safe=False, then an existing value will be overwritten.

        Parameters
        ----------------

        name : str
            The parameter name that will be used to identify it

        val : <anything>
            An object or primitive we wish to associate with this attribute

        safe : bool (default = True)
            Flag which if True with throw an exception if an attribute with 
            the same name already exists, otherwise the newly introduced 
            attribute will overwrite the previous one.

        Returns
        ---------
            None - but adds an attribute to the calling object

        """

        if safe:
            if name in self._attributes:
                raise SiteException("Trying to add attribute [%s=%s] to Site [%s] but this attribute is already set.\nPossible options are: %s" %(name,val, str(self), str(self._attributes.keys())))
                
        self._attributes[name] = val


    ## ------------------------------------------------------------------------
    ##
    def remove_attribute(self, name, safe=True):
        """
        Function that removes a given attribute from the Site based on the 
        passed attribute name. If the passed attribute does not exist or is not 
        associate with the Site then this will trigger an exception 
        unless safe=False.

        Parameters
        ----------------

        name : str
            The attribute name that will be used to identify it

        safe : bool (default = True)
            Flag which if True with throw an exception if an 
            attribute this name does not exists. If set to
            False then if an attribute is not found it is simply
            ignored
            
        Returns
        ---------
        None
            No return type but will remove an attribute from the 
            protein if present.
            
        """

        if name not in self._attributes:
            if safe:
                raise ProteinException(f'Passed attribute [{name}] not found in {self}')
        else:
            del self._attributes[name]
        
        

    ## ------------------------------------------------------------------------
    ##
    def get_local_sequence_context(self, offset = 5):
        """
        Returns the local amino acid context around a residue +/- the offset
        provided.
        
        Note that the offset extends to the start/end of sequence and then 
        silently truncates.
        
        Parameters
        -----------
        offset : int
            Defines the +/- region around the position which is used to 
            define the local sequence context.
            

        Returns
        --------
        str
            Returns an amino acid sequence that corresponds to the local 
            sequence context around the site of interest
        """

        return self._protein.get_sequence_context(self.position, offset)    



    #######################################
    ##                                   ##
    ##        SITE domain FUNCTIONS      ##
    ##                                   ##
    #######################################

    ## ------------------------------------------------------------------------
    ##      
    def get_domains(self, offset=0, safe=True):
        """
        Function that returns the set of domains that the site lies within. The
        oofset parameter defines the wiggle room +/- that is tolerated, but
        defaults to 0.         

        Parameters
        --------------

        offset : int (default = 0)
            +/- values around the site from which regions are taken.

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, else they 
            just return None.

        Returns
        ----------
        list
            Returns a list of domain objects for which this site can be found
            in or near

        """
        
        valid_domains = []

        for domain in self.protein.domains:

            # get start and end positions with offset
            (p1, p2) = sequence_utilities.get_bounding_sites(domain.start, offset, self._protein._len)
            start = p1
            
            (p1, p2) = sequence_utilities.get_bounding_sites(domain.end, offset, self._protein._len)
            end = p2

            if start <= self.position and end >= self.position:
                valid_domains.append(domain)

        return valid_domains



    #######################################
    ##                                   ##
    ##        SITE TRACK FUNCTIONS       ##
    ##                                   ##
    #######################################

    ## ------------------------------------------------------------------------
    ##      
    def get_track_values(self, name, offset=0, safe=True):
        """
        Function that returns the region of a protein's values- track 
        associated with this site, +/- some offset.
        
        If the track name is missing and safe is True, this will throw 
        an exception, otherwise (if safe=False) then if the track is 
        missing the function returns None.


        Parameters
        --------------
        
        name : str
            Track name

        offset : int (default = 0)
            +/- values around the site from which regions are taken

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, else 
            they just return None


        Returns
        ----------
        list
            Returns a list of floats that corresponds to the set of 
            residues associated with the domain of interest
            

        """
        
        (p1, p2) = sequence_utilities.get_bounding_sites(self._position, offset, self._protein._len)
        

        # because calling values_region only makes sense IF the track exists, we have to split
        # this into two operations. Note this throws an exception if safe=True and the track
        # does not exist
        t = self._protein.track(name, safe)

        if t is not None:
            return t.values_region(p1, p2)
        else:
            return None


    ## ------------------------------------------------------------------------
    ##      
    def get_track_value(self, name, safe=True):
        """
        Function that returns the value associated with the track at
        the residue position associated with this site.

        If the track name is missing and safe is True, this will throw 
        an exception, otherwise (if safe=False) then if the track is 
        missing the function returns None.


        Parameters
        --------------
        
        name : str
            Track name

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, else 
            they just return None

        Returns
        ----------
        float or int
            Returns the value associated with the track of interest
            at this site.

        """
        return self.get_track_values(name, offset=0, safe=safe)[0]


    ## ------------------------------------------------------------------------
    ##      
    def get_track_symbols(self, name, offset=0, safe=True):
        """
        Function that returns the region of a protein's symbols track 
        associated with this site.
                
        If a Track of this name is not associated with the underlying
        protein and safe is True, this will throw an exception, 
        otherwise (if safe=False) then if the track is missing the 
        function returns None.

        Parameters
        --------------
        
        name : str
            Track name

        offset : int (default = 0)
            +/- values around the site from which regions are taken

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, 
            else they just return None


        Returns
        ----------
        list
            Returns a list of strs that corresponds to the set of 
            residues associated with the domain of interest.
        """

        (p1, p2) = sequence_utilities.get_bounding_sites(self._position, offset, self._protein._len)

        # because calling symbols_region only makes sense IF the track exists, we have to split
        # this into two operations. Note this throws an exception if safe=True and the
        # track does not exist
        t = self._protein.track(name, safe)

        if t is not None:
            return t.symbols_region(p1,p2)
        else:
            return None


    ## ------------------------------------------------------------------------
    ##      
    def get_track_symbol(self, name, safe=True):
        """
        Function that returns the symbol associated with the track at
        the residue position associated with this site.
                
        If a Track of this name is not associated with the underlying
        protein and safe is True, this will throw an exception, 
        otherwise (if safe=False) then if the track is missing the 
        function returns None.

        Parameters
        --------------
        
        name : str
            Track name

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, 
            else they just return None

        Returns
        ----------
        str
            Returns the string associated with the symbol at this site

        """

        return self.get_track_symbols(name, offset=0, safe=safe)[0]


    ## ------------------------------------------------------------------------
    ##
    def __repr__(self):             
        return "|Site: %s @ %i in protein %s" % (self._site_type, self.position, self.protein.unique_ID)
    
    



# must import pytest
import pytest

# import additional modules needed for this code
import shephard
from shephard.apis import uniprot  
from shephard import interfaces

# import the domain_tools module which we're going to test
from shephard.tools import domain_tools

def test_first_test():

    # this line uses shephards internal get_data function to build the full
    # path to the directory where our test data is
    test_data_dir = shephard.get_data('test_data')

    # define filenames - note we use the names we defined above
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    domain_file = '%s/%s' % (test_data_dir, 'TS1_domains_idr.tsv')

    # create a new Proteome object and annotate with the domains
    # in the domain_file
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_domains.add_domains_from_file(P, domain_file)

    P.protein('O00401').domains

    assert domain_tools.calculate_domain_length(P.protein('O00401').domains[0]) == 20




"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder
Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)
Holehouse Lab - Washington University in St. Louis
"""

import pytest
from shephard.exceptions import ProteinException, DomainException
import numpy as np
import shephard
from shephard.apis import uniprot
from shephard import interfaces
from shephard.tools import domain_tools

names_list = ['O00401', 'O00470', 'O00472', 'O00499', 'O00629', 'O00712', 'O00716', 'O14786', 'Q9UJX3']

def test_domain_overlap():    
    """
    Note if any of these succeed the test suite will fail because it incrases
    """

    # read in a proteome
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # add a synthetic protein and extract it
    P.add_protein('A'*100,'test','test')    
    prot = P.protein('test')

    
    prot.add_domain(1,  20, 'test_domain_0') # 0
    prot.add_domain(5,  15, 'test_domain_1') # 1

    d0 = prot.get_domains_by_type('test_domain_0')[0]
    d1 = prot.get_domains_by_type('test_domain_1')[0]
    assert domain_tools.domain_overlap(d0, d1) is True
    assert domain_tools.domain_overlap(d1, d0) is True


    prot.add_domain(15, 25, 'test_domain_2') # 2
    d2 = prot.get_domains_by_type('test_domain_2')[0]
    assert domain_tools.domain_overlap(d0, d2) is True
    assert domain_tools.domain_overlap(d1, d2) is True
    assert domain_tools.domain_overlap(d2, d0) is True
    assert domain_tools.domain_overlap(d2, d1) is True

    prot.add_domain(20, 30, 'test_domain_3') # 3
    d3 = prot.get_domains_by_type('test_domain_3')[0]
    assert domain_tools.domain_overlap(d0, d3) is True
    assert domain_tools.domain_overlap(d2, d3) is True
    assert domain_tools.domain_overlap(d1, d3) is False

    assert domain_tools.domain_overlap(d3, d0) is True
    assert domain_tools.domain_overlap(d3, d2) is True
    assert domain_tools.domain_overlap(d3, d1) is False

    prot.add_domain(1, 4, 'test_domain_4') # 4
    d4 = prot.get_domains_by_type('test_domain_4')[0]
    assert domain_tools.domain_overlap(d0, d4) is True
    assert domain_tools.domain_overlap(d1, d4) is False

    assert domain_tools.domain_overlap(d4, d0) is True
    assert domain_tools.domain_overlap(d4, d1) is False

    print(prot.domains)
    prot.add_domain(1, 5, 'test_domain_5') # 5
    d5 = prot.get_domains_by_type('test_domain_5')[0]
    assert domain_tools.domain_overlap(d0, d5) is True
    assert domain_tools.domain_overlap(d1, d5) is True

    assert domain_tools.domain_overlap(d5, d0) is True
    assert domain_tools.domain_overlap(d5, d1) is True

def test_domain_overlap_fraction():    
    """
    """

    
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    

    # TODO write a series of tests that test the domain_overlap_fraction
    # functionality

    P.add_protein('A'*1000,'test','test')    
    prot = P.protein('test')

    ## TEST 1
    prot.add_domain(1,100,'test1')
    prot.add_domain(1,100,'test2')

    d1 = prot.domains[0]
    d2 = prot.domains[1]
    

    assert domain_tools.domain_overlap_fraction(d1,d2) == 1.0
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0


    ## TEST 2
    prot.add_domain(1,100,'test1')
    prot.add_domain(50,100,'test2')

    d1 = prot.domains[0]
    d2 = prot.domains[1]
    

    assert domain_tools.domain_overlap_fraction(d1,d2) == 1.0
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0



    ## TEST 3
    prot.add_domain(1,100,'test1')
    prot.add_domain(101,120,'test2')

    d1 = prot.domains[0]
    d2 = prot.domains[1]
    

    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.0
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0



    ## TEST 3.1 (switched order of addition    
    prot.add_domain(101,120,'test2')
    prot.add_domain(1,100,'test1')


    d1 = prot.domains[0]
    d2 = prot.domains[1]
    

    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.0
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0


    ## TEST 4
    prot.add_domain(1,100,'test1')
    prot.add_domain(81,120,'test2')
    


    d1 = prot.domains[0]
    d2 = prot.domains[1]
    
    print(len(d1))
    print(len(d2))

    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.5
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0


    ## TEST 5
    # here only 1 residue overlaps for test2 (residue 100)
    # and the domain is 100-109 (i..e 10 residues) SOOOO 
    # the fraction overlap should be 0.1
    prot.add_domain(1,100,'test1')
    prot.add_domain(100,109,'test2')
    


    d1 = prot.domains[0]
    d2 = prot.domains[1]
    
    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.1
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0



    ## TEST 5.1
    # here only 1 residue overlaps for test2 (residue 100)
    # and the domain is 100-109 (i..e 10 residues) SOOOO 
    # the fraction overlap should be 0.1
    # same as above but change order of addition
    
    prot.add_domain(100,109,'test1')
    prot.add_domain(1,100,'test2')


    d1 = prot.domains[0]
    d2 = prot.domains[1]

    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.1
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0


    ## TEST 6
    # here only 1 residue overlaps for test2 (residue 109)
    # and the domain is 100-109 (i..e 10 residues) SOOOO 
    # the fraction overlap should be 0.1
    prot.add_domain(109,200,'test1')
    prot.add_domain(100,109,'test2')
    


    d1 = prot.domains[0]
    d2 = prot.domains[1]
    
    assert domain_tools.domain_overlap_fraction(d1,d2) == 0.1
    prot.remove_domain(d1)
    prot.remove_domain(d2)
    assert len(prot.domains) == 0
    
    ## Test 7.x - cyle over all possible x/10 overlaps
    for i in range(109,99,-1):
        
        prot.add_domain(i,200,'test1')
        prot.add_domain(100,109,'test2')

        prediction = (110-i)/10

        d1 = prot.domains[0]
        d2 = prot.domains[1]
    
        assert domain_tools.domain_overlap_fraction(d1,d2) == prediction
        prot.remove_domain(d1)
        prot.remove_domain(d2)
        assert len(prot.domains) == 0


    ## Test 8.x - cyle over all possible x/10 overlaps
    for i in range(100,109,1):
        
        prot.add_domain(1,i,'test1')
        prot.add_domain(100,109,'test2')

        prediction = (1+(i-100))/10

        d1 = prot.domains[0]
        d2 = prot.domains[1]
    
        assert domain_tools.domain_overlap_fraction(d1,d2) == prediction
        prot.remove_domain(d1)
        prot.remove_domain(d2)
        assert len(prot.domains) == 0





"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import pytest
import numpy as np

from shephard.tools import sequence_tools

def test_build_mega_string(TS1):
    """
    This function tests the global concatination of sequences 
    from either protien or domain objects
    """
    
    # test passing of protien objects 
    assert sequence_tools.build_mega_string(TS1) == ''.join([i.sequence for i in TS1])
    
    # test passing of domain objects 
    assert sequence_tools.build_mega_string(TS1.domains) == ''.join([i.sequence for i in TS1.domains])

def test_find_string_positions():
    """
    This function tests the global concatination of sequences 
    from either protiens or domains
    """
    s0 = 'AACABBBBACABA'
    s1 = "A"
    s2 = 'CAB'
    s3 = 'Z'
    
    # test multible positions 
    assert sequence_tools.find_string_positions(s1, s0) == [1,2,4,9,11,13]
    
    # test multible characters 
    assert sequence_tools.find_string_positions(s2, s0) == [3,10]
    
    # test nothing 
    assert sequence_tools.find_string_positions(s3, s0) == []


import pytest

from shephard.exceptions import ProteinException
from shephard.exceptions import InterfaceException
import shephard
from shephard.interfaces import si_sites
from shephard.apis import uniprot  


test_data_dir = shephard.get_data('test_data')


def test_si_site_add_from_file():

    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites.tsv'))

    O00470 = TS1.protein('O00470')

    # check length
    assert len(O00470) == 390
    assert len(O00470.site(1)) == 1
    assert len(O00470.site(390)) == 1

    # get site a position 2 that doesn't exist
    with pytest.raises(ProteinException) as e_info:
        O00470.site(2)

    # get site at position 2 that doesn't exist
    assert O00470.site(2, safe=False) == []


def test_si_site_add_from_file_test_robustness():

    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    # this should fail - skip_bad influences skipping
    with pytest.raises(ProteinException) as e_info:
        si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites_bad.tsv'), skip_bad=True)

    with pytest.raises(ProteinException) as e_info:
        si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites_bad.tsv'), skip_bad=False)

    # but if we load in with safe=False will skip over the bad site
    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites_bad.tsv'), safe=False)
    O00470 = TS1.protein('O00470')

    assert len(O00470) == 390



        
def test_si_site_add_file_read_robustness():

    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    # thsi should fail - malformatted line
    with pytest.raises(InterfaceException) as e_info:    
        si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites_bad2.tsv'), skip_bad=False)


    si_sites.add_sites_from_file(TS1, '%s/%s' % (test_data_dir, 'ts1_bonus_sites_bad2.tsv'), skip_bad=True)
    assert len(TS1.sites) == 4


def test_write_sites():

    
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    for p in TS1:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['S','P','E']:
                p.add_site(s, 'test_site')

    si_sites.write_sites(TS1, 'output_test/test_sites.tsv')


    TS2 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))
    si_sites.add_sites_from_file(TS2, 'output_test/test_sites.tsv')

    for p in TS2:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['S','P','E']:
                assert len(p.site(s)) == 1



def test_write_sites_with_attributes():

    
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    for p in TS1:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['S','P','E']:
                p.add_site(s, 'test_site', symbol='A', value=0.2, attributes={'test_at_1':0, 'test_at_2':'yeah', 'test_at_3':1/33})

    si_sites.write_sites(TS1, 'output_test/test_sites.tsv')


    TS2 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))
    si_sites.add_sites_from_file(TS2, 'output_test/test_sites.tsv')

    for p in TS2:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['S','P','E']:
                assert len(p.site(s)) == 1
                assert p.site(s)[0].attribute('test_at_1') == '0'
                assert p.site(s)[0].attribute('test_at_2') == 'yeah'
                assert p.site(s)[0].attribute('test_at_3') == str(1/33)
                assert p.site(s)[0].symbol == 'A'
                assert p.site(s)[0].value == 0.2


def test_write_sites_from_list():

    
    TS1 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))

    for p in TS1:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['M','A','K']:
                p.add_site(s, 'test_site', symbol='A', value=0.2, attributes={'test_at_1':0, 'test_at_2':'yeah', 'test_at_3':1/33})

    all_sites = TS1.sites
    
    si_sites.write_sites_from_list(all_sites, 'output_test/test_sites.tsv')


    TS2 = uniprot.uniprot_fasta_to_proteome('%s/%s' % (test_data_dir,'testset_1.fasta'))
    si_sites.add_sites_from_file(TS2, 'output_test/test_sites.tsv')

    for p in TS2:
        
        for s in range(1,len(p)+1):
            if p.residue(s) in ['M','A','K']:
                assert len(p.site(s)) == 1
                assert p.site(s)[0].attribute('test_at_1') == '0'
                assert p.site(s)[0].attribute('test_at_2') == 'yeah'
                assert p.site(s)[0].attribute('test_at_3') == str(1/33)
                assert p.site(s)[0].symbol == 'A'
                assert p.site(s)[0].value == 0.2


            
            
